from __future__ import annotations

from dataclasses import dataclass
from functools import wraps
from typing import TYPE_CHECKING, Any, cast

from memory_profiler import memory_usage

if TYPE_CHECKING:
    from collections.abc import Callable


def memory_profiled[**P, T](func: Callable[P, T], /) -> Callable[P, Output[T]]:
    """Call a function, but also profile its maximum memory usage."""

    @wraps(func)
    def wrapped(*args: P.args, **kwargs: P.kwargs) -> Output[T]:
        memory, value = memory_usage(
            cast("Any", (func, args, kwargs)), max_usage=True, retval=True
        )
        return Output(value=value, memory=memory)

    return wrapped


@dataclass(kw_only=True, slots=True)
class Output[T]:
    """A function output, and its memory usage."""

    value: T
    memory: float


__all__ = ["Output", "memory_profiled"]



from __future__ import annotations

import datetime as dt
from typing import cast

from lightweight_charts import Chart
from polars import DataFrame, Date, Float64, Int64, col
from pytest import fixture, raises

from utilities.lightweight_charts import (
    _SetDataFrameEmptyError,
    _SetDataFrameNonUniqueError,
    set_dataframe,
)


@fixture
def df() -> DataFrame:
    data = [
        (dt.date(2000, 1, 1), 5824.0, 5824.75, 5823.75, 5824.25, 124),
        (dt.date(2000, 1, 2), 5821.0, 5821.5, 5820.75, 5820.75, 146),
        (dt.date(2000, 1, 3), 5820.0, 5821.0, 5819.75, 5820.5, 128),
        (dt.date(2000, 1, 4), 5822.5, 5822.5, 5822.25, 5822.25, 78),
        (dt.date(2000, 1, 5), 5822.5, 5822.5, 5821.5, 5821.75, 73),
        (dt.date(2000, 1, 6), 5817.0, 5817.0, 5816.0, 5816.5, 301),
        (dt.date(2000, 1, 7), 5817.75, 5818.75, 5817.5, 5818.75, 150),
        (dt.date(2000, 1, 8), 5821.0, 5821.25, 5821.0, 5821.25, 75),
        (dt.date(2000, 1, 9), 5818.0, 5819.0, 5818.0, 5818.75, 69),
        (dt.date(2000, 1, 10), 5818.75, 5819.25, 5818.5, 5819.0, 67),
    ]
    return DataFrame(
        data=data,
        schema={
            "date": Date,
            "open": Float64,
            "high": Float64,
            "low": Float64,
            "close": Float64,
            "volume": Int64,
        },
        orient="row",
    )


class TestSetDataFrame:
    def test_main(self, *, df: DataFrame) -> None:
        chart = Chart()
        set_dataframe(df, chart)

    def test_error_empty(self, *, df: DataFrame) -> None:
        df = df.drop("date")
        with raises(
            _SetDataFrameEmptyError,
            match="At least 1 column must be of date/datetime type; got 0",
        ):
            set_dataframe(df, cast("Chart", None))

    def test_error_non_unique(self, *, df: DataFrame) -> None:
        df = df.with_columns(col("date").alias("date2"))
        with raises(
            _SetDataFrameNonUniqueError,
            match=r"Schema\(.*\) must contain exactly 1 date/datetime column; got 'date', 'date2' and perhaps more",
        ):
            set_dataframe(df, cast("Chart", None))



from __future__ import annotations

from collections.abc import Callable
from functools import partial
from pathlib import Path
from typing import TYPE_CHECKING, Any, cast, overload

import atools
from whenever import TimeDelta

from utilities.types import Coro, PathLike

if TYPE_CHECKING:
    from atools._memoize_decorator import Keygen, Pickler


type _Key[**P, T] = tuple[Callable[P, Coro[T]], TimeDelta]
_MEMOIZED_FUNCS: dict[_Key, Callable[..., Coro[Any]]] = {}


async def call_memoized[**P, T](
    func: Callable[P, Coro[T]],
    refresh: TimeDelta | None = None,
    /,
    *args: P.args,
    **kwargs: P.kwargs,
) -> T:
    """Call an asynchronous function, with possible memoization."""
    if refresh is None:
        return await func(*args, **kwargs)
    key: _Key = (func, refresh)
    memoized_func: Callable[P, Coro[T]]
    try:
        memoized_func = _MEMOIZED_FUNCS[key]
    except KeyError:
        memoized_func = _MEMOIZED_FUNCS[(key)] = memoize(duration=refresh.in_seconds())(
            func
        )
    return await memoized_func(*args, **kwargs)


##


@overload
def memoize[F: Callable[..., Coro[Any]]](
    func: F,
    /,
    *,
    db_path: PathLike | None = None,
    duration: float | TimeDelta | None = None,
    keygen: Keygen | None = None,
    pickler: Pickler | None = None,
    size: int | None = None,
) -> F: ...
@overload
def memoize[F: Callable[..., Coro[Any]]](
    func: None = None,
    /,
    *,
    db_path: PathLike | None = None,
    duration: float | TimeDelta | None = None,
    keygen: Keygen | None = None,
    pickler: Pickler | None = None,
    size: int | None = None,
) -> Callable[[F], F]: ...
def memoize[F: Callable[..., Coro[Any]]](
    func: F | None = None,
    /,
    *,
    db_path: PathLike | None = None,
    duration: float | TimeDelta | None = None,
    keygen: Keygen | None = None,
    pickler: Pickler | None = None,
    size: int | None = None,
) -> F | Callable[[F], F]:
    """Memoize a function."""
    if func is None:
        result = partial(
            memoize,
            db_path=db_path,
            duration=duration,
            keygen=keygen,
            pickler=pickler,
            size=size,
        )
        return cast("Callable[[F], F]", result)
    return atools.memoize(
        db_path=None if db_path is None else Path(db_path),
        duration=duration.py_timedelta()
        if isinstance(duration, TimeDelta)
        else duration,
        keygen=keygen,
        pickler=pickler,
        size=size,
    )(func)


__all__ = ["call_memoized"]



from __future__ import annotations

from inspect import signature
from pathlib import Path
from time import sleep
from typing import TYPE_CHECKING, ClassVar

from pytest import fixture, mark, param, raises

from utilities.iterables import one
from utilities.os import temp_environ
from utilities.pytest import NodeIdToPathError, node_id_path, throttle

if TYPE_CHECKING:
    from collections.abc import Sequence

    from _pytest.legacypath import Testdir


@fixture(autouse=True)
def set_asyncio_default_fixture_loop_scope(*, testdir: Testdir) -> None:
    _ = testdir.makepyprojecttoml("""
        [tool.pytest.ini_options]
        asyncio_default_fixture_loop_scope = "function"
    """)


class TestNodeIdPath:
    @mark.parametrize(
        ("node_id", "expected"),
        [
            param(
                "src/tests/module/test_funcs.py::TestClass::test_main",
                Path("src.tests.module.test_funcs/TestClass__test_main"),
            ),
            param(
                "src/tests/module/test_funcs.py::TestClass::test_main[param1, param2]",
                Path(
                    "src.tests.module.test_funcs/TestClass__test_main[param1, param2]"
                ),
            ),
            param(
                "src/tests/module/test_funcs.py::TestClass::test_main[EUR.USD]",
                Path("src.tests.module.test_funcs/TestClass__test_main[EUR.USD]"),
            ),
        ],
    )
    def test_main(self, *, node_id: str, expected: Path) -> None:
        result = node_id_path(node_id)
        assert result == expected

    @mark.parametrize(
        "node_id",
        [
            param("src/tests/module/test_funcs.py::TestClass::test_main"),
            param(
                "python/package/src/tests/module/test_funcs.py::TestClass::test_main"
            ),
        ],
    )
    def test_root(self, *, node_id: str) -> None:
        result = node_id_path(node_id, root="src/tests")
        expected = Path("module.test_funcs/TestClass__test_main")
        assert result == expected

    def test_suffix(self) -> None:
        node_id = "src/tests/module/test_funcs.py::TestClass::test_main"
        result = node_id_path(node_id, root="src/tests", suffix=".csv")
        expected = Path("module.test_funcs/TestClass__test_main.csv")
        assert result == expected

    def test_error_file_suffix(self) -> None:
        with raises(NodeIdToPathError, match="Node ID must be a Python file; got .*"):
            _ = node_id_path("src/tests/module/test_funcs.csv::TestClass::test_main")


class TestPytestOptions:
    def test_unknown_mark(self, *, testdir: Testdir) -> None:
        _ = testdir.makepyfile(
            """
            from pytest import mark

            @mark.unknown
            def test_main() -> None:
                assert True
            """
        )
        result = testdir.runpytest()
        result.assert_outcomes(errors=1)
        result.stdout.re_match_lines([r".*Unknown pytest\.mark\.unknown"])

    @mark.parametrize("configure", [param(True), param(False)])
    def test_unknown_option(self, *, configure: bool, testdir: Testdir) -> None:
        if configure:
            _ = testdir.makeconftest(
                """
                from utilities.pytest import add_pytest_configure

                def pytest_configure(config):
                    add_pytest_configure(config, [("slow", "slow to run")])
                """
            )
        _ = testdir.makepyfile(
            """
            def test_main() -> None:
                assert True
            """
        )
        result = testdir.runpytest("--unknown")
        result.stderr.re_match_lines([r".*unrecognized arguments.*"])

    @mark.parametrize(
        ("case", "passed", "skipped", "matches"),
        [param([], 0, 1, [".*3: pass --slow"]), param(["--slow"], 1, 0, [])],
    )
    def test_one_mark_and_option(
        self,
        *,
        testdir: Testdir,
        case: Sequence[str],
        passed: int,
        skipped: int,
        matches: Sequence[str],
    ) -> None:
        _ = testdir.makeconftest(
            """
            from utilities.pytest import add_pytest_addoption
            from utilities.pytest import add_pytest_collection_modifyitems
            from utilities.pytest import add_pytest_configure

            def pytest_addoption(parser):
                add_pytest_addoption(parser, ["slow"])

            def pytest_collection_modifyitems(config, items):
                add_pytest_collection_modifyitems(config, items, ["slow"])

            def pytest_configure(config):
                add_pytest_configure(config, [("slow", "slow to run")])
            """
        )
        _ = testdir.makepyfile(
            """
            from pytest import mark

            @mark.slow
            def test_main() -> None:
                assert True
            """
        )
        result = testdir.runpytest("-rs", *case)
        result.assert_outcomes(passed=passed, skipped=skipped)
        result.stdout.re_match_lines(list(matches))

    @mark.parametrize(
        ("case", "passed", "skipped", "matches"),
        [
            param(
                [],
                1,
                3,
                [
                    "SKIPPED.*: pass --slow",
                    "SKIPPED.*: pass --fast",
                    "SKIPPED.*: pass --slow --fast",
                ],
            ),
            param(
                ["--slow"],
                2,
                2,
                ["SKIPPED.*: pass --fast", "SKIPPED.*: pass --slow --fast"],
            ),
            param(
                ["--fast"],
                2,
                2,
                ["SKIPPED.*: pass --slow", "SKIPPED.*: pass --slow --fast"],
            ),
            param(["--slow", "--fast"], 4, 0, []),
        ],
    )
    def test_two_marks_and_options(
        self,
        *,
        testdir: Testdir,
        case: Sequence[str],
        passed: int,
        skipped: int,
        matches: Sequence[str],
    ) -> None:
        _ = testdir.makeconftest(
            """
            from utilities.pytest import add_pytest_addoption
            from utilities.pytest import add_pytest_collection_modifyitems
            from utilities.pytest import add_pytest_configure

            def pytest_addoption(parser):
                add_pytest_addoption(parser, ["slow", "fast"])

            def pytest_collection_modifyitems(config, items):
                add_pytest_collection_modifyitems(
                    config, items, ["slow", "fast"],
                )

            def pytest_configure(config):
                add_pytest_configure(
                    config, [("slow", "slow to run"), ("fast", "fast to run")],
                )
            """
        )
        _ = testdir.makepyfile(
            """
            from pytest import mark

            def test_none() -> None:
                assert True

            @mark.slow
            def test_slow() -> None:
                assert True

            @mark.fast
            def test_fast() -> None:
                assert True

            @mark.slow
            @mark.fast
            def test_both() -> None:
                assert True
            """
        )
        result = testdir.runpytest("-rs", *case, "--randomly-dont-reorganize")
        result.assert_outcomes(passed=passed, skipped=skipped)
        result.stdout.re_match_lines(list(matches))


class TestThrottle:
    delta: ClassVar[float] = 0.5

    @mark.flaky
    @mark.parametrize("on_try", [param(True), param(False)])
    def test_basic(self, *, testdir: Testdir, tmp_path: Path, on_try: bool) -> None:
        _ = testdir.makepyfile(
            f"""
            from whenever import TimeDelta

            from utilities.pytest import throttle

            @throttle(root={str(tmp_path)!r}, delta=TimeDelta(seconds={self.delta}), on_try={on_try})
            def test_main() -> None:
                assert True
            """
        )
        testdir.runpytest().assert_outcomes(passed=1)
        testdir.runpytest().assert_outcomes(skipped=1)
        sleep(self.delta)
        testdir.runpytest().assert_outcomes(passed=1)

    @mark.flaky
    @mark.parametrize("asyncio_first", [param(True), param(False)])
    @mark.parametrize("on_try", [param(True), param(False)])
    def test_async(
        self, *, testdir: Testdir, tmp_path: Path, asyncio_first: bool, on_try: bool
    ) -> None:
        if asyncio_first:
            _ = testdir.makepyfile(
                f"""
                from whenever import TimeDelta

                from pytest import mark

                from utilities.pytest import throttle

                @mark.asyncio
                @throttle(root={str(tmp_path)!r}, delta=TimeDelta(seconds={self.delta}), on_try={on_try})
                async def test_main() -> None:
                    assert True
                """
            )
        else:
            _ = testdir.makepyfile(
                f"""
                from whenever import TimeDelta

                from pytest import mark

                from utilities.pytest import throttle

                @throttle(root={str(tmp_path)!r}, delta=TimeDelta(seconds={self.delta}), on_try={on_try})
                @mark.asyncio
                async def test_main() -> None:
                    assert True
                """
            )
        testdir.runpytest().assert_outcomes(passed=1)
        testdir.runpytest().assert_outcomes(skipped=1)
        sleep(self.delta)
        testdir.runpytest().assert_outcomes(passed=1)

    @mark.flaky
    @mark.parametrize("on_try", [param(True), param(False)])
    def test_disabled_via_env_var(
        self, *, testdir: Testdir, tmp_path: Path, on_try: bool
    ) -> None:
        _ = testdir.makepyfile(
            f"""
            from whenever import TimeDelta

            from utilities.pytest import throttle

            @throttle(root={str(tmp_path)!r}, delta=TimeDelta(seconds={self.delta}), on_try={on_try})
            def test_main() -> None:
                assert True
            """
        )
        with temp_environ(THROTTLE="1"):
            testdir.runpytest().assert_outcomes(passed=1)
            testdir.runpytest().assert_outcomes(passed=1)
            sleep(self.delta)
            testdir.runpytest().assert_outcomes(passed=1)

    @mark.flaky
    def test_on_pass(self, *, testdir: Testdir, tmp_path: Path) -> None:
        _ = testdir.makeconftest(
            """
            from pytest import fixture

            def pytest_addoption(parser):
                parser.addoption("--pass", action="store_true")

            @fixture
            def is_pass(request) -> bool:
                return request.config.getoption("--pass")
            """
        )
        _ = testdir.makepyfile(
            f"""
            from whenever import TimeDelta

            from utilities.pytest import throttle

            @throttle(root={str(tmp_path)!r}, delta=TimeDelta(seconds={self.delta}))
            def test_main(*, is_pass: bool) -> None:
                assert is_pass
            """
        )
        for delta_use in [self.delta, 0.0]:
            testdir.runpytest().assert_outcomes(failed=1)
            testdir.runpytest("--pass").assert_outcomes(passed=1)
            testdir.runpytest("--pass").assert_outcomes(skipped=1)
            sleep(delta_use)

    @mark.flaky
    def test_on_try(self, *, testdir: Testdir, tmp_path: Path) -> None:
        _ = testdir.makeconftest(
            """
            from pytest import fixture

            def pytest_addoption(parser):
                parser.addoption("--pass", action="store_true")

            @fixture
            def is_pass(request):
                return request.config.getoption("--pass")
            """
        )
        root_str = str(tmp_path)
        _ = testdir.makepyfile(
            f"""
            from whenever import TimeDelta

            from utilities.pytest import throttle

            @throttle(root={root_str!r}, delta=TimeDelta(seconds={self.delta}), on_try=True)
            def test_main(*, is_pass: bool) -> None:
                assert is_pass
            """
        )
        for delta_use in [self.delta, 0.0]:
            testdir.runpytest().assert_outcomes(failed=1)
            testdir.runpytest().assert_outcomes(skipped=1)
            sleep(self.delta)
            testdir.runpytest("--pass").assert_outcomes(passed=1)
            testdir.runpytest().assert_outcomes(skipped=1)
            sleep(delta_use)

    @mark.flaky
    def test_long_name(self, *, testdir: Testdir, tmp_path: Path) -> None:
        _ = testdir.makepyfile(
            f"""
            from pytest import mark
            from string import printable
            from whenever import TimeDelta

            from utilities.pytest import throttle

            @mark.parametrize("arg", [10 * printable])
            @throttle(root={str(tmp_path)!r}, delta=TimeDelta(seconds={self.delta}))
            def test_main(*, arg: str) -> None:
                assert True
            """
        )
        testdir.runpytest().assert_outcomes(passed=1)
        testdir.runpytest().assert_outcomes(skipped=1)
        sleep(self.delta)
        testdir.runpytest().assert_outcomes(passed=1)

    def test_signature(self) -> None:
        @throttle()
        def func(*, fix: bool) -> None:
            assert fix

        def other(*, fix: bool) -> None:
            assert fix

        assert signature(func) == signature(other)

    @mark.flaky
    def test_error_decoding_timestamp(
        self, *, testdir: Testdir, tmp_path: Path
    ) -> None:
        _ = testdir.makepyfile(
            f"""
            from whenever import TimeDelta

            from utilities.pytest import throttle

            @throttle(root={str(tmp_path)!r}, delta=TimeDelta(seconds={self.delta}))
            def test_main() -> None:
                assert True
            """
        )
        testdir.runpytest().assert_outcomes(passed=1)
        path = one(tmp_path.iterdir())
        _ = path.write_text("invalid")
        testdir.runpytest().assert_outcomes(passed=1)



from __future__ import annotations

from hashlib import md5
from typing import Any


def md5_hash(obj: Any, /) -> str:
    """Compute the MD5 hash of an arbitrary object."""
    from utilities.orjson import serialize

    return md5(serialize(obj), usedforsecurity=False).hexdigest()


__all__ = ["md5_hash"]



from __future__ import annotations

from os import environ
from pathlib import Path
from string import Template
from typing import Any, assert_never


def substitute_environ(path_or_text: Path | str, /, **kwargs: Any) -> str:
    """Substitute the environment variables in a file."""
    match path_or_text:
        case Path() as path:
            return substitute_environ(path.read_text(), **kwargs)
        case str() as text:
            return Template(text).substitute(environ, **kwargs)
        case never:
            assert_never(never)


__all__ = ["substitute_environ"]



from __future__ import annotations

from itertools import chain
from typing import TYPE_CHECKING

from hypothesis import given
from hypothesis.strategies import (
    DataObject,
    booleans,
    data,
    integers,
    just,
    lists,
    none,
    sampled_from,
    sets,
)
from pytest import mark, param, raises

from utilities.hypothesis import sentinels, text_ascii
from utilities.text import (
    ParseBoolError,
    ParseNoneError,
    _SplitKeyValuePairsDuplicateKeysError,
    _SplitKeyValuePairsSplitError,
    _SplitStrClosingBracketMismatchedError,
    _SplitStrClosingBracketUnmatchedError,
    _SplitStrCountError,
    _SplitStrOpeningBracketUnmatchedError,
    join_strs,
    parse_bool,
    parse_none,
    repr_encode,
    secret_str,
    snake_case,
    split_key_value_pairs,
    split_str,
    str_encode,
    strip_and_dedent,
    to_bool,
    to_str,
    unique_str,
)

if TYPE_CHECKING:
    from collections.abc import Sequence

    from utilities.sentinel import Sentinel


class TestParseBool:
    @given(data=data(), value=booleans())
    def test_main(self, *, data: DataObject, value: bool) -> None:
        match value:
            case True:
                extra_cased_texts = ["Y", "Yes", "On"]
            case False:
                extra_cased_texts = ["N", "No", "Off"]
        all_cased_texts = list(chain([str(value), str(int(value))], extra_cased_texts))
        all_texts = list(
            chain(
                extra_cased_texts,
                map(str.lower, all_cased_texts),
                map(str.upper, all_cased_texts),
            )
        )
        text = data.draw(sampled_from(all_texts))
        result = parse_bool(text)
        assert result is value

    @given(
        text=sampled_from([
            "00",
            "11",
            "ffalsee",
            "invalid",
            "nn",
            "nnoo",
            "oofff",
            "oonn",
            "ttruee",
            "yy",
            "yyess",
        ])
    )
    def test_error(self, *, text: str) -> None:
        with raises(ParseBoolError, match="Unable to parse boolean value; got '.*'"):
            _ = parse_bool(text)


class TestParseNone:
    @given(data=data())
    def test_main(self, *, data: DataObject) -> None:
        text = str(None)
        text_use = data.draw(sampled_from(["", text, text.lower(), text.upper()]))
        result = parse_none(text_use)
        assert result is None

    @given(text=sampled_from(["invalid", "nnonee"]))
    def test_error(self, *, text: str) -> None:
        with raises(ParseNoneError, match="Unable to parse null value; got '.*'"):
            _ = parse_none(text)


class TestReprEncode:
    @given(n=integers())
    def test_main(self, *, n: int) -> None:
        result = repr_encode(n)
        expected = repr(n).encode()
        assert result == expected


class TestSecretStr:
    def test_main(self) -> None:
        s = secret_str("text")
        assert repr(s) == secret_str._REPR
        assert str(s) == secret_str._REPR

    def test_open(self) -> None:
        s = secret_str("text")
        assert isinstance(s.str, str)
        assert not isinstance(s.str, secret_str)
        assert repr(s.str) == repr("text")
        assert str(s.str) == "text"


class TestSnakeCase:
    @given(
        case=sampled_from([
            ("API", "api"),
            ("APIResponse", "api_response"),
            ("ApplicationController", "application_controller"),
            ("Area51Controller", "area51_controller"),
            ("FreeBSD", "free_bsd"),
            ("HTML", "html"),
            ("HTMLTidy", "html_tidy"),
            ("HTMLTidyGenerator", "html_tidy_generator"),
            ("HTMLVersion", "html_version"),
            ("NoHTML", "no_html"),
            ("One   Two", "one_two"),
            ("One  Two", "one_two"),
            ("One Two", "one_two"),
            ("OneTwo", "one_two"),
            ("One_Two", "one_two"),
            ("One__Two", "one_two"),
            ("One___Two", "one_two"),
            ("Product", "product"),
            ("SpecialGuest", "special_guest"),
            ("Text", "text"),
            ("Text123", "text123"),
            ("_APIResponse_", "_api_response_"),
            ("_API_", "_api_"),
            ("__APIResponse__", "_api_response_"),
            ("__API__", "_api_"),
            ("__impliedVolatility_", "_implied_volatility_"),
            ("_itemID", "_item_id"),
            ("_lastPrice__", "_last_price_"),
            ("_symbol", "_symbol"),
            ("aB", "a_b"),
            ("changePct", "change_pct"),
            ("changePct_", "change_pct_"),
            ("impliedVolatility", "implied_volatility"),
            ("lastPrice", "last_price"),
            ("memMB", "mem_mb"),
            ("sizeX", "size_x"),
            ("symbol", "symbol"),
            ("testNTest", "test_n_test"),
            ("text", "text"),
            ("text123", "text123"),
        ])
    )
    def test_main(self, *, case: tuple[str, str]) -> None:
        text, expected = case
        result = snake_case(text)
        assert result == expected


class TestSplitKeyValuePairs:
    @given(
        case=sampled_from([
            ("", []),
            ("a=1", [("a", "1")]),
            ("a=1,b=22", [("a", "1"), ("b", "22")]),
            ("a=1,b=22,c=333", [("a", "1"), ("b", "22"), ("c", "333")]),
            ("=1", [("", "1")]),
            ("a=", [("a", "")]),
            ("a=1,=22,c=333", [("a", "1"), ("", "22"), ("c", "333")]),
            ("a=1,b=,c=333", [("a", "1"), ("b", ""), ("c", "333")]),
            ("a=1,b=(22,22,22),c=333", [("a", "1"), ("b", "(22,22,22)"), ("c", "333")]),
            ("a=1,b=(c=22),c=333", [("a", "1"), ("b", "(c=22)"), ("c", "333")]),
        ])
    )
    def test_main(self, *, case: tuple[str, Sequence[tuple[str, str]]]) -> None:
        text, expected = case
        result = split_key_value_pairs(text)
        assert result == expected

    def test_mapping(self) -> None:
        result = split_key_value_pairs("a=1,b=22,c=333", mapping=True)
        expected = {"a": "1", "b": "22", "c": "333"}
        assert result == expected

    def test_error_split_list(self) -> None:
        with raises(
            _SplitKeyValuePairsSplitError,
            match=r"Unable to split 'a=1,b=\(c=22\],d=333' into key-value pairs",
        ):
            _ = split_key_value_pairs("a=1,b=(c=22],d=333")

    def test_error_split_pair(self) -> None:
        with raises(
            _SplitKeyValuePairsSplitError,
            match=r"Unable to split 'a=1,b=22=22,c=333' into key-value pairs",
        ):
            _ = split_key_value_pairs("a=1,b=22=22,c=333")

    def test_error_duplicate_keys(self) -> None:
        with raises(
            _SplitKeyValuePairsDuplicateKeysError,
            match=r"Unable to split 'a=1,a=22,a=333' into a mapping since there are duplicate keys; got \{'a': 3\}",
        ):
            _ = split_key_value_pairs("a=1,a=22,a=333", mapping=True)


class TestSplitAndJoinStr:
    @given(
        data=data(),
        case=sampled_from([
            ("", 0, []),
            (r"\,", 1, [""]),
            (",", 2, ["", ""]),
            (",,", 3, ["", "", ""]),
            ("1", 1, ["1"]),
            ("1,22", 2, ["1", "22"]),
            ("1,22,333", 3, ["1", "22", "333"]),
            ("1,,333", 3, ["1", "", "333"]),
            ("1,(22,22,22),333", 5, ["1", "(22", "22", "22)", "333"]),
        ]),
    )
    def test_main(self, *, data: DataObject, case: tuple[str, int, list[str]]) -> None:
        text, n, expected = case
        n_use = data.draw(just(n) | none())
        result = split_str(text, n=n_use)
        if n_use is None:
            assert result == expected
        else:
            assert result == tuple(expected)
        assert join_strs(result) == text

    @given(
        data=data(),
        case=sampled_from([
            ("1", 1, ["1"]),
            ("1,22", 2, ["1", "22"]),
            ("1,22,333", 3, ["1", "22", "333"]),
            ("1,(22),333", 3, ["1", "(22)", "333"]),
            ("1,(22,22),333", 3, ["1", "(22,22)", "333"]),
            ("1,(22,22,22),333", 3, ["1", "(22,22,22)", "333"]),
        ]),
    )
    def test_brackets(
        self, *, data: DataObject, case: tuple[str, int, list[str]]
    ) -> None:
        text, n, expected = case
        n_use = data.draw(just(n) | none())
        result = split_str(text, brackets=[("(", ")")], n=n_use)
        if n_use is None:
            assert result == expected
        else:
            assert result == tuple(expected)
        assert join_strs(result) == text

    @given(texts=lists(text_ascii()))
    def test_generic(self, *, texts: list[str]) -> None:
        assert split_str(join_strs(texts)) == texts

    @given(texts=sets(text_ascii()))
    def test_sort(self, *, texts: set[str]) -> None:
        assert split_str(join_strs(texts, sort=True)) == sorted(texts)

    def test_error_closing_bracket_mismatched(self) -> None:
        with raises(
            _SplitStrClosingBracketMismatchedError,
            match=r"Unable to split '1,\(22\},333'; got mismatched '\(' at position 2 and '}' at position 5",
        ):
            _ = split_str("1,(22},333", brackets=[("(", ")"), ("{", "}")])

    def test_error_closing_bracket_unmatched(self) -> None:
        with raises(
            _SplitStrClosingBracketUnmatchedError,
            match=r"Unable to split '1,22\),333'; got unmatched '\)' at position 4",
        ):
            _ = split_str("1,22),333", brackets=[("(", ")")])

    def test_error_count(self) -> None:
        with raises(
            _SplitStrCountError,
            match=r"Unable to split '1,22,333' into 4 part\(s\); got 3",
        ):
            _ = split_str("1,22,333", n=4)

    def test_error_opening_bracket(self) -> None:
        with raises(
            _SplitStrOpeningBracketUnmatchedError,
            match=r"Unable to split '1,\(22,333'; got unmatched '\(' at position 2",
        ):
            _ = split_str("1,(22,333", brackets=[("(", ")")])


class TestStrEncode:
    @given(n=integers())
    def test_main(self, *, n: int) -> None:
        result = str_encode(n)
        expected = str(n).encode()
        assert result == expected


class TestStripAndDedent:
    @mark.parametrize("trailing", [param(True), param(False)])
    def test_main(self, *, trailing: bool) -> None:
        text = """
               This is line 1.
               This is line 2.
               """
        result = strip_and_dedent(text, trailing=trailing)
        expected = "This is line 1.\nThis is line 2." + ("\n" if trailing else "")
        assert result == expected


class TestToBool:
    @given(bool_=booleans() | none() | sentinels())
    def test_bool_none_or_sentinel(self, *, bool_: bool | None | Sentinel) -> None:
        assert to_bool(bool_) is bool_

    @given(bool_=booleans())
    def test_str(self, *, bool_: bool) -> None:
        assert to_bool(str(bool_)) is bool_

    @given(bool_=booleans())
    def test_callable(self, *, bool_: bool) -> None:
        assert to_bool(lambda: bool_) is bool_


class TestToStr:
    @given(text=text_ascii())
    def test_str(self, *, text: str) -> None:
        assert to_str(text) == text

    @given(text=text_ascii())
    def test_callable(self, *, text: str) -> None:
        assert to_str(lambda: text) == text

    @given(text=none() | sentinels())
    def test_none_or_sentinel(self, *, text: None | Sentinel) -> None:
        assert to_str(text) is text


class TestUniqueStrs:
    def test_main(self) -> None:
        first, second = [unique_str() for _ in range(2)]
        assert first != second



from __future__ import annotations

from asyncio import Task, create_task
from typing import TYPE_CHECKING, Any, override

from fastapi import FastAPI
from uvicorn import Config, Server

from utilities.asyncio import timeout_td
from utilities.contextlib import enhanced_async_context_manager
from utilities.whenever import get_now_local

if TYPE_CHECKING:
    from collections.abc import AsyncIterator

    from utilities.types import Delta, MaybeType


_TASKS: list[Task[None]] = []


class _PingerReceiverApp(FastAPI):
    """App for the ping pinger."""

    @override
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)  # skipif-ci

        @self.get("/ping")  # skipif-ci
        def ping() -> str:
            return f"pong @ {get_now_local()}"  # skipif-ci

        _ = ping  # skipif-ci


@enhanced_async_context_manager
async def yield_ping_receiver(
    port: int,
    /,
    *,
    host: str = "localhost",
    timeout: Delta | None = None,
    error: MaybeType[BaseException] = TimeoutError,
) -> AsyncIterator[None]:
    """Yield the ping receiver."""
    app = _PingerReceiverApp()  # skipif-ci
    server = Server(Config(app, host=host, port=port))  # skipif-ci
    _TASKS.append(create_task(server.serve()))  # skipif-ci
    try:  # skipif-ci
        async with timeout_td(timeout, error=error):
            yield
    finally:  # skipif-ci
        await server.shutdown()


__all__ = ["yield_ping_receiver"]
